/* 
   contentManager.hh

   Description see below.
 
     Copyright (C) 2005, Patrick Rammelt <rammelt@cs.tu-berlin.de>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
*/

/* ========================================================================
   Content Management 
   ( template base class )

    _________________________ C O N D I T I O N S _____________________________
   /__________________________________________________________________________/
                   
   A class inheriting from contentManager<ContType> (e.g. T) should 
   encapsulate all its attributes (the "content") in a separate class which is
   given in the template parameter ContType. This class (say C) must have at 
   least a default and a copy-constructor. The content can be accessed via the
   methods contentManager<C>::cont() and contentManager<C>::read() - the 
   former returns a non-const / the latter a const pointer to the content
   Scetch:
		      class C { 
		      protected:
		          // ... all attributes ...
		      public:
		          C (void);               // default constructor
		          C (const class C &t);   // copy-constructor
		          // ... element functions ...
	              };

		      class T : public contentManager<C> {
		      protected:
		          // ... _no_ attributes ...
		      public:
		          // ... element functions ...
                      };

    __________________________ A S S I G N I N G_______________________________
   /__________________________________________________________________________/

   Assigning one object to another ("x=y") is done by calling
   x.contentManager<C>::connect(y), which just sets the content pointer to the 
   very same content already referenced by y. Furthermore x and y (and all 
   other objects assigned to the same content) will be connected in a
   "reference-copy-chain".
   e.g.: 
		      class T : public contentManager<C> {
		      public:
		          // calling connect:
		          T (const class T &t) { 
                              connect (t); 
	      	          }
		          class T& operator = (const class T &t) {
		      	      connect (t); 
			  }
                      };

   Notice: connect(t) for const t is allowed because copy/reference links
           are declared "mutable" (i.e. non-const even for const objects)

    ___________________ R E A D I N G   A T T R I B U T E S ___________________
   /__________________________________________________________________________/

   Unless a content is explicetely created there just is none, i.e. 
   contentManager<C>::cont() / contentManager<C>::read() gives a NULL-pointer 
   (no content). To ensure to have a content before trying to read attribute-
   values from it call contentManager<C>::validate(). This will create a new 
   content (from "scratch") by calling the default constructor.
   e.g.: 
		      class T : public contentManager<C> {
                      public:
			  // default constructor
		          T (void) {
		              //             ... do nothing ...
		      	      // ...(at least do not create a content)...
			  }
			  // check if content exists
                          void readContent (void) {
			      if (read()==NULL)
			          // ... no content yet - bummer ...
			      else 
			          // ... reading from content ...
		          }
			  // make sure there is a content
		          void reallyReadContent (void) {
		              validate();
			      // ... reading from content ...
                          }
		      };

   Notice  I: Do not create a content in the default constructor, because this
              will be called by all constructors of derived classes (e.g. a 
	      copy constructor using connect() should not create a "private" 
	      content before connecting).
   Notice II: Methods called by a (default) constructor cannot be overwritten 
              even when declared virtual (i.e.: you cannot change the behavior
	      of a constructor from inheriting classes)

    __________________ C H A N G I N G   A T T R I B U T E S __________________
   /__________________________________________________________________________/

   If an object is about to be changed (normally by its own member-functions 
   only) this got to be "notified" by a call to contentManager<C>::recreate() 
   (create new content from scratch) or contentManager<C>::write() (create a 
   "private" copy of the current content). 
   In any case those objects then referencing to their "own" content and being
   "checked-out" of the reference-copy-chain i.e being "alone" - as long as an
   object (re-)connects to it. Attempts to connect can be blocked/unblocked by 
   calling contentManager<C>::lock()/unlock(). Objects trying to connect to a 
   locked object are enforced to make a "hard-copy".
   e.g.: 
		      class T : public contentManager<C> {
                      public:
		          // changing values
			  void changeMyself (void) { 
			      write();
			      lock();   // do not connect to me
			      // ... do whatever has to be done ...
			      unlock(); // ready
		          }
			  // create a complete new content anyway
		          void recreateMyself (void) {
			      recreate(); 
			      lock();
			      // ... fill content ...
			      unlock();
		          }
			  // recreate but need old values for that
		          void recreateMyself2 (void) {
			      class T t(*this); // make a ("soft") copy
			      recreate();       // notify recreation
			      lock();
			      // ... fill content (using t) ...
			      unlock();
		          }
			  // do not need a content anymore (why?)
		          void getRidOfContent (void) {
			      deconnect(); 
			      // ( never ever use: delete[] cont();  )
		          }
		      };

   Notice: Lock counts "lock-levels", i.e. after: lock(); lock(); unlock();
           object is still locked.

    _________ P A S S I N G   O B J E C T S   T O   F U N C T I O N S _________
   /__________________________________________________________________________/

   If you would have trouble with passing "this" as a parameter to a method,
   use call by value type parameter passing.
   Remember the copy-constructor of class T calls connect(), which avoids 
   unnecessary copying. For internal purposes on the other hand - when we can 
   ensure to not use "*this" - one should use reference calls for performance 
   reasons.
   e.g.
		      class T : public contentManager<C> {
                      protected: // ... or even private
	      	          void m1 (const class T& t) {
		              // being sure &t!=this
		          }
		      public: // ... methods for the wide world
			  void m3 (class T t) {
			      // call by value avoids all problems
		              // with "t==this"
			  }
                      };


    ______ R E T U R N I N G   O B J E C T S   F O M   F U N C T I O N S ______
   /__________________________________________________________________________/

   Just do it like this:

                      class T<C> foo () {
		          class T<C> result; 
			  // fill result
			  return (result);
                      }

   CopyManager avoids unnecessary copying when calling x = foo();

    ___________________________ R E C U R S I O N _____________________________
   /__________________________________________________________________________/

   It is perfectly ok to use attributes in a content class of a class managed 
   by contentManager which are managed by contentManager itself (sounds 
   complicated? - see below ...)
   e.g.:
		      // we know that class - don't we?
		      class T : public contentManager<C> { ... };

		      // content class for class S (see below)
		      class K { 
		      protected:
		          class T t1;    // t1 is also a copyManger-object
		          class T t2;    // ...
		      public:
		          // copy constructor
		          K (class K &k) { 
 		              t1 = k.t1; // guess what happens here ...
		              t2 = k.t2; // ( [1] see below )
			  }
			  // copy constructor for const argument
			  // (notice: all attributes of k are also const)
			  K (const class K &k) {
			      t1 = k.t1; // this will do "hard"-copying
			      t2 = k.t2;
			  }
		      };

		      class S : public contentManager<K> { 
			  // ... all as described for class T ...
		      };

		  [1] notifying a chanche in t2 will deconnect only t2 from 
                      k.t2, while t1 and k.t1 still sharing the same content 

    __________________________ I N H E R I T A N C E __________________________
   /__________________________________________________________________________/

   The major drawback: it is impossible to add attributes when inheriting from
   a class T which is "managed" by contentManager unless class T is a template 
   class which takes the content-type as a template parameter (just like class
   contentManager)
   e.g.:
                      // "content" class
		      class C { ... };

		      // "main" class managed by contentManager
		      template <class ContType>
		      class T : public contentManager<ContType> { ... };
			   
		      // advanced "content" class
		      class C2 : public C { ... };

		      // advanced "main" class
		      class T2 : public T<C2> { ... };

		      int main (void) {
		         class T<C>   simple;   // using simple version
		         class T2<C2> advanced; // using advanced version
		      }


    _________________________ A L T E R N A T I V E S _________________________
   /__________________________________________________________________________/

   Depending on the special conditions of an application, the use of the 
   auto_ptr class might be an alternative to return objects from functions.
   The return value then is a (smart) pointer (which is imho less elegant 
   compared to return by value), but the user of the function at least don't 
   have to worry about freeing memory for the referenced object.

   Newer compilers also implement the "named return value optimization" (nrvo)
   which also avoids the copying of an object returned by value from a 
   function.
   
   Nevertheless both methods do not offer an alternative for having equal 
   objects without really copying them unless it is necessary. As a consequnce
   they also doesn't have the advantage that objects can be passed by value 
   into methods in order get a copy e.g. when passing the object itself into 
   its own method ( X.foo(X) ).

   ------------------------------------------------------------------------
   Author: Patrick Rammelt
   Date:   25.08.2004
   ======================================================================== */

#ifndef _ETC_CONTENT_MANAGER_HH
#define _ETC_CONTENT_MANAGER_HH


#include <stdio.h>
#include <stdlib.h>


namespace ContentManager {

  
/* ========================================================================
   Copy Management - Class definition
   ======================================================================== */

template <class ContType>
class contentManager {

private:

  // -----------------------------< attributes >--------------------------- //

  /* Content */
  ContType *cnt;

  /* Reference to the object which contains the data (initially: ref==this) */
  mutable const class contentManager *ref; 

  /* Reference to the object whichs ref is this (initially: cpy==this) */
  mutable const class contentManager *cpy;

  /* Locked counter - while an object is locked (lck>0) no one can 
     connect ( got to make a "hard"-copy instead ) */
  int lck;
 
public:

  // -----------------------< construct / destruct >----------------------- //

  /* construct
     notice: content is NULL */
  contentManager (void) : cnt(NULL), ref(this), cpy(this), lck(0) {}

  /* destruct
     this deletes content only if we are alone (no ref-cpy-chain) */
  virtual ~contentManager (void) { deconnect(); }

  /* copy constructors 
     just to get the idea - this has to be done also in inheriting classes */
  contentManager (const class contentManager &obj) 
    : cnt(NULL), ref(this), cpy(this), lck(0) { connect(obj); }

  /* assigning one object to another 
     just to get the idea - this has to be done also in inheriting classes */
  virtual class contentManager& operator = (const class contentManager &obj) { 
    connect(obj); return (*this); 
  }

public:

  // ---------------------------< locked / ... >--------------------------- //

  /* lock / unlock / total unlock */
  inline void lock   (void) { lck++; }
  inline void unlock (void) { if (lck>0) lck--; }
  inline void delock (void) { lck = 0; }

  /* is this locked? / lock-level (i.e. number 
     of calls to unlock() until !isLocked() */
  inline bool isLocked (void) { return (lck>0); }
  inline int lockLevel (void) { return (lck); }

  /* check to be "alone" ( i.e. the content is just for me ) */
  inline bool isMine (void) { return (ref==this && cpy==this); }

protected:

  // ---------------------------< copy-management >------------------------ //

  /* make this equal to obj
     put this into the ref-cpy-chain of obj and set cnt=obj.cnt */
  void connect (const class contentManager &obj);

  /* notify for deletion 
     eventually delete content if we are alone) */
  void deconnect (void);

  /* get read-only-access to current content (always up to date) 
     this might return NULL if there is no content */
  inline const ContType* read (void) const { return (cnt); }

  /* get access to current content (always up-to-date)
     change content only after a call to recreate(), write() or validate(true),
     otherwise you are in danger to change more than one object */
  inline ContType* cont (void) { return (cnt); }

  /* notify a write-access and get content 
     content will be created if there is none
     use up-to-date cont() for further access to content */
  ContType* write (void);

  /* notify a recreation and get content 
     content will be created if there is none
     use up-to-date cont() for further access to content */
  ContType* recreate (void);

public:

  /* create an (own copy of the) content if there is none or chkout==true */
  inline void validate (bool chkout = 1) { if (chkout || cnt==NULL) write(); }

private:

  // -------------------------------< helpers >---------------------------- //

  /* checks out of the ref-cpy-chain
     but does not change content cnt */
  inline void checkout (void);

  /* delete content cnt
     use this only if we are "alone" i.e. ref==this && cpy==this */
  inline void del_content  (void) { delete cnt; cnt = NULL; } 
  
  /* - creates a copy of cont and store it in cnt 
     - creates new content from scratch if cont==NULL
     - do not care of old cnt (eventually one needs to delete old cnt first) */
  inline void copy_content (const ContType *cont) { 
    if (cont==NULL) cnt = new ContType;      // create new cont from scratch
    else cnt = new ContType(*cont);          // create new copy of cont
  }
    
};


/* ========================================================================
   Copy Management - Implementation of Member-Functions
   ======================================================================== */

//              _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ //
// CONNECT "=" _/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/ //
//                                                                          //

/* ------------------------------------------------------------------------
   make this equal to obj
   ------------------------------------------------------------------------ */
template <class ContType>
void contentManager<ContType>::connect (const contentManager<ContType> &obj) 
{
  if (obj.cnt != cnt) {          // we do not have the same content

    deconnect();                 // check out and eventually delete content

    if (lck) {
      copy_content (obj.cnt);    // it's locked ... must do a "hard" copy
    } else {
      cnt = obj.cnt;             // now we have the same content
      ref = &obj;                // our ref is obj (make this equal to obj)
      if (obj.cpy!=&obj) {       // obj got a copy (there is a "chain")
	cpy = obj.cpy;           // put us between obj and obj.cpy
	cpy->ref = this;
      }
      obj.cpy = this;            // we are the new copy of obj
    }

  }
}


//                        _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ //
// NOTIFY CONTENT-ACCESS _/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/ //
//                                                                          //


/* ------------------------------------------------------------------------
   notify a write-access and get content
   ------------------------------------------------------------------------ */
template <class ContType>
ContType* contentManager<ContType>::write (void)
{
  if (ref!=this || cpy!=this) {   // we are not alone
    copy_content (cnt);           // we need our own copy now
    checkout();                   // check out of the ref-cpy-chain
  } else if (cnt==NULL) {         // we are alone and without a content
    cnt = new ContType;           // create new content
  }
  return (cnt);
}  


/* ------------------------------------------------------------------------
   notify a recreation and get content
   ------------------------------------------------------------------------ */
template <class ContType>
ContType* contentManager<ContType>::recreate (void) 
{
  if (ref!=this || cpy!=this) {   // we are not alone
    cnt = new ContType;           // create new content just for us
    checkout();                   // check out of the ref-cpy-chain
  } else if (cnt==NULL) {         // we are alone and without a content
    cnt = new ContType;           // create new content
  }
  return (cnt);
}  


/* ------------------------------------------------------------------------
   notify for deletion / delete  content
   ------------------------------------------------------------------------ */
template <class ContType> 
void contentManager<ContType>::deconnect (void) 
{
  if (ref==this && cpy==this) {   // we are alone
    del_content();                // it's our job to free memory
  } else {
    cnt = NULL;                   // just leave content for the others
    checkout();                   // check out of the ref-cpy-chain
  }
}  


//                  _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ //
// "PRIVATE PARTS" _/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/ //
//                                                                          //

/* ------------------------------------------------------------------------
   check out of the ref-cpy-chain
   ------------------------------------------------------------------------ */
template <class ContType>
inline void contentManager<ContType>::checkout (void) 
{
  if (cpy!=this) {                // there is a object pointing to us

    if (ref!=this) {              // we are pointing to another object
      cpy->ref = ref;             // set reference of copy to our reference
      ref->cpy = cpy;             // now our copy is the copy of ref
      ref      = this;            // we are alone now ...
      cpy      = this;
    } else {                      // we are containing the data 
      cpy->ref = cpy;             // now (hopefully) copy contains our data
      cpy      = this;            // we are alone now ...
    }

  } else if (ref!=this) {         // we are pointing to another object

    ref->cpy = ref;               // we are no longer the copy of ref
    ref      = this;              // we are alone now ...

  }
}


// ------------------------------------------------------------------------ //


} // namespace ContentManager


#endif // ifndef _ETC_CONTENT_MANAGER_HH
